"""
Customized matplotlib artist primitives
"""

import matplotlib as mpl
from matplotlib.lines import Line2D
import numpy as np
import warnings

from typing import Optional, Any, Union, Literal, Collection, Dict, Tuple

from .utils import deep_update

affine = mpl.transforms.Affine2D


class EnergyLevel(Line2D):
    """
    Energy level artist.

    This object also implements a number of potential Text artists for labelling.
    It also includes helper methods for getting the exact coordinates of anchor points
    for connected coupling arrows and the like.
    """

    def __str__(self):
        if self._energy is None:
            return "EnergyLevel()"
        else:
            return "EnergyLevel((%g,%g))" % (self._xpos, self._energy)

    def __init__(
        self,
        energy: float,
        xpos: float,
        width: float,
        right_text: str = "",
        left_text: str = "",
        top_text: str = "",
        bottom_text: str = "",
        text_kw: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        """
        Parameters:
            energy (float): y-axis position of the level
            xpos (float): x-axis position of the level
            width (float): Width of the level line, in units of the x-axis
            right_text (str, optional): Text to put to the right of the level
            left_text (str, optional): Text to put to the left of the level
            top_text (str, optional): Text to put above the level
            bottom_text (str, optional): Text to put below the level
            text_kw (dict, optional): Dictionary of keyword-arguments passed to
                :class:`matplotlib:matplotlib.text.Text`
            kwargs: Passed to the :class:`matplotlib:matplotlib.lines.Line2D` constructor
        """

        self._energy = energy
        self._xpos = xpos
        self._width = width

        if text_kw is None:
            text_kw = {}
        # we'll update the position when the line data is set
        self.text_labels: Dict[str, mpl.text.Text] = {
            "right": mpl.text.Text(
                xpos + width / 2, energy, right_text, ha="left", va="center", **text_kw
            ),
            "left": mpl.text.Text(
                xpos - width / 2, energy, left_text, ha="right", va="center", **text_kw
            ),
            "top": mpl.text.Text(
                xpos, energy, top_text, ha="center", va="bottom", **text_kw
            ),
            "bottom": mpl.text.Text(
                xpos, energy, bottom_text, ha="center", va="top", **text_kw
            ),
        }
        """Text label objects to add to the level"""

        x = (xpos - width / 2, xpos + width / 2)
        y = (energy, energy)

        super().__init__(x, y, **kwargs)

    def get_center(self) -> np.ndarray:
        """
        Returns coordinates of the center of the level line.

        Returns:
            numpy.ndarray: x,y coordinates
        """

        return np.array((self._xpos, self._energy), dtype=float)

    def get_left(self) -> np.ndarray:
        """
        Returns coordinates of the left of the level line.

        Returns:
            numpy.ndarray: x,y coordinates
        """

        return np.array((self._xpos - self._width / 2, self._energy), dtype=float)

    def get_right(self) -> np.ndarray:
        """
        Returns coordinates of the right of the level line.

        Returns:
            numpy.ndarray: x,y coordinates
        """

        return np.array((self._xpos + self._width / 2, self._energy), dtype=float)

    def get_anchor(
        self, loc: Union[Literal["center", "left", "right"], Collection] = "center"
    ) -> np.ndarray:
        """
        Returns an anchor on the level in plot coordinates.

        Parameters:
            loc (str or collection of 2 elements): What reference point to return.
                `'center'`, `'left'`, `'right'` gives those points of the level.
                A 2-element iterable is interpreted as offsets from the center
                location.

        Raises:
            TypeError: If `loc` is not accepted string or a 2-element iterable.
        """

        if loc == "center":
            anchor = self.get_center()
        elif loc == "left":
            anchor = self.get_left()
        elif loc == "right":
            anchor = self.get_right()
        else:
            if len(loc) == 2:
                anchor = self.get_center() + np.array(loc)
            else:
                raise TypeError("loc must iterable of two elements if not using keys")

        return anchor

    def set_figure(self, figure):
        for side, label in self.text_labels.items():
            label.set_figure(figure)
        super().set_figure(figure)

    def set_axes(self, axes: mpl.axes.Axes):
        for side, label in self.text_labels.items():
            label.set_axes(axes)
        super().set_axes(axes)

    def set_transform(self, transform):
        """
        Overridden to add padding offsets to labels.
        """
        # pixel offsets
        pad = 6
        offsets = {
            "right": (pad, 0),
            "left": (-pad, 0),
            "top": (0, pad),
            "bottom": (0, -pad),
        }
        for side, label in self.text_labels.items():
            label.set_transform(transform + affine().translate(*offsets[side]))
        super().set_transform(transform)

    def set_data(self, x, y):

        super().set_data(x, y)

    def draw(self, renderer):

        super().draw(renderer)
        for side, label in self.text_labels.items():
            label.draw(renderer)


class Coupling(Line2D):
    """
    Coupling artist for showing couplings between levels.

    This artist is a conglomeration of artists.

    - :class:`Line2D <matplotlib:matplotlib.lines.Line2D>` for the actual coupling path
    - :class:`Polygon <matplotlib:matplotlib.patches.Polygon>` for the arrow heads
    - :class:`Text <matplotlib:matplotlib.text.Text>` for the label

    Sufficient methods are overridden from the base Line2D class to ensure
    the other artists are rendered whenever the main artist is rendered.
    """

    def __str__(self) -> str:

        return "Coupling((%g,%g)->(%g,%g))" % (*self._start, *self._stop)

    def __init__(
        self,
        start: Collection,
        stop: Collection,
        arrowsize: float,
        deflection: float = 0.0,
        waveamp: float = 0.0,
        halfperiod: float = 0,
        arrowratio: float = 1,
        tail: bool = False,
        arrow_kw: Optional[Dict[str, Any]] = None,
        label: str = "",
        label_offset: Literal["center", "left", "right"] = "center",
        label_rot: bool = False,
        label_flip: bool = False,
        label_kw: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        """
        Parameters:
            start (2-element collection): Coupling start location in data coordinates
            stop (2-element collection): Coupling end location in data coordinates
            arrowsize (float): Size of arrowheads in x-data coordinates
            deflection (float, optional): Amount to bend the center of the coupling arrow
                away from linear.
                Defined in y-coordintes. Default is 0 or no deflection.
            waveamp (float, optional): Amplitude of sine wave modulation of the coupling arrow.
                Defined in y-coordinates. Default is 0 or no waviness.
            halfperiod (float, optional): Length of half-period of sine modulation.
                Defined in x-coordinates.
                Both `waveamp` and `period` must be non-zero to get modulation.
                Default is 0 or no waviness.
            arrowratio (float, optional): Aspect ratio of the arrowhead.
                Default is 1 for equal aspect ratio.
            tail (bool, optional): Whether to draw an identical arrowhead at the coupling base.
                Default is False.
            arrow_kw (dict, optional): Dictionary of keyword arguments to pass to
                :class:`matplotlib:matplotlib.patches.Polygon` constructor.
                Note that keyword arguments provided to this function will clobber
                identical keys provided here.
            label (str, optional): Label string to apply to the coupling.
                Default is no label.
            label_offset (str, optional): Offset direction for the label.
                Options are `'center'`, `'left'`, and `'right'`.
                Default is center of the coupling line.
            label_rot (bool, optional): Label is justified along the coupling arrow axis if True.
                Default is False, so label is oriented along x-axis always.
            label_flip (bool, optional): Apply a 180 degree rotation to the label.
                Default is False.
            label_kw (dict, optional): Dictionary of keyword arguments to pass to the
                :class:`matplotlib:matplotlib.text.Text` constructor.
            kwargs: Optional keyword arguments passed to the
                :class:`matplotlib:matplotlib.lines.Line2D` constructor and the
                :class:`matplotlib:matplotlib.patches.Polygon` constructor for the arrowhead.
                Note that `'color'` will be automatically changed to `'facecolor'` for
                the arrowhead to avoid extra lines.
        """

        if not isinstance(start, Collection) or not isinstance(stop, Collection):
            raise RuntimeError("x/y data must be a sequence of two elements")
        elif len(start) != 2 or len(stop) != 2:
            raise RuntimeError("x/y data must be a sequence of two elements")

        if arrow_kw is None:
            arrow_kw = {}
        if label_kw is None:
            label_kw = {}

        self._start = np.array(start)
        self._stop = np.array(stop)
        self._arrowsize = arrowsize
        self._arrowratio = arrowratio
        self._tail = tail
        self.tail = False

        self._deflection = deflection
        self._deflect = not np.isclose(deflection, 0.0)

        self._waveamp = waveamp
        self._halfperiod = halfperiod

        # set number of points to use when drawing lines
        self.npts = 100

        if arrowsize * arrowratio < waveamp:
            warnings.warn(
                "Wave amplitude is larger than arrowhead; result will look funny."
            )

        x0, y0 = self.init_path()
        super().__init__(x0, y0, **kwargs)

        # initialize arrowheads
        # use line kwargs to set arrow defaults
        arrow_kw.update(kwargs)
        # use facecolor instead of color
        color = arrow_kw.pop("color", None)
        if color is not None:
            arrow_kw["fc"] = color
        self.init_arrowheads(**arrow_kw)

        # initialize label text
        self.init_label(label, label_offset, label_rot, label_flip, **label_kw)

    def init_path(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculates the desired path for the line of the coupling.

        The returned path is a line relative to the x-axis of the correct length.
        Transforms are used to move and rotate this path to the end location.
        This method of making the couplings is a little convoluted,
        but it allows for simple definition of very general paths (line sine waves)
        without distortions.

        Returns
        -------
        x: numpy.ndarray
            x-coordinates of the data points for the un-rotated, un-translated path
        y: numpy.ndarray
            y-coordinates of the data points for the un-rotated, un-translated path
        """

        vec = self._stop - self._start
        dist = np.sqrt(vec.dot(vec))
        self._dist = dist
        self._ang = np.arctan2(*vec[::-1])

        h = self._deflection

        # convert half-periods
        try:
            omega = np.pi / self._halfperiod
        except ZeroDivisionError:
            omega = 0.0
        
        if self._deflect:

            if h > dist/2:
                warnings.warn("Maximum deflection exceded "
                              "(deflection > half vector length). "
                              "Behavior poorly defined.")
        
            r = (dist**2+4*h**2)/(8*h)
            theta_center = 2*np.arcsin(dist/2/r)
            # accounts for shortened path from arrowheads
            theta_shim = 2*np.arcsin(self._arrowsize/2/r)
            phi = theta_center/2
            self._phi = phi
            # accounts for shortened path when rotating arrowheads
            self._phi_shim = phi - theta_shim/2

            if self._tail:
                thetas = np.linspace(theta_shim, theta_center-theta_shim, self.npts)
                theta0 = theta_center-theta_shim*2
            else:
                thetas = np.linspace(0, theta_center - theta_shim, self.npts)
                theta0 = theta_center-theta_shim

            xc = ((r+self._waveamp*
                   np.sin(omega*r*theta_center*thetas/theta0))
                  * np.sin(thetas-phi)
                  + r*np.sin(phi))
            yc = ((r+self._waveamp*
                  np.sin(omega*r*theta_center*thetas/theta0))
                  * np.cos(thetas-phi)
                  - r*np.cos(phi))
        
            return xc, yc
        
        else:
            
            if self._tail:
                x0 = np.linspace(self._arrowsize, dist - self._arrowsize, self.npts)
                op0 = omega * self._arrowsize*2
            else:
                x0 = np.linspace(0, dist - self._arrowsize, self.npts)
                op0 = omega * self._arrowsize
                
            y0 = self._waveamp * np.sin(omega*x0 + op0)
                
            return x0, y0

    def init_arrowheads(self, **kwargs):
        """
        Creates the arrowhead(s) for the coupling as matplotlib polygon objects.

        Parameters:
            kwargs: Optional keyword arguments to pass to the
                :class:`matplotlib:matplotlib.patches.Polygon` constructor.
        """

        verts0 = np.array([[-1, 0.5], [-1, -0.5], [0, 0], [-1, 0.5]])
        # set aspect ratio of arrow
        verts0[:,1] *= self._arrowratio
        
        if self._deflect:
            verts_head = verts0 @ self._rotation_matrix(self._phi_shim)
            self.head = mpl.patches.Polygon(verts_head, closed=False, **kwargs)
            if self._tail:
                verts_tail = verts0 @ self._rotation_matrix(-self._phi_shim)
                self.tail = mpl.patches.Polygon(verts_tail, closed=False, **kwargs)
                
        else:
            self.head = mpl.patches.Polygon(verts0, closed=False, **kwargs)
            
            if self._tail:
                self.tail = mpl.patches.Polygon(verts0, closed=False, **kwargs)

    def _rotation_matrix(self, theta):
        
        rot= np.array([[np.cos(theta), -np.sin(theta)],
                       [np.sin(theta), np.cos(theta)]])
        
        return rot

    def init_label(
        self,
        label: str,
        label_offset: Literal["center", "right", "left"],
        label_rot: bool,
        label_flip: bool,
        **label_kw
    ):
        """
        Creates the coupling label text object.

        Parameters:
            label (str): Label string to apply to the coupling.
            label_offset (str): Offset direction for the label.
                Options are `'center'`, `'left'`, and `'right'`.
            label_rot (bool): Label will be justified along the coupling arrow axis if True.
            label_flip (bool): Apply a 180 degree rotation to the label.
            label_kw: Keyword arguments to pass to the
                :class:`matplotlib:matplotlib.text.Text` constructor.
        """

        # define default softbackground for text
        bbox_defaults = {
            "boxstyle": "round,pad=0.05",
            "fc": "w",
            "ec": "none",
            "alpha": 0.5,
        }
        bbox_dict = label_kw.pop("bbox", {})
        bbox_dict = deep_update(bbox_defaults, bbox_dict)
        label_kw["bbox"] = bbox_dict

        if label_offset == "center":
            label_ha = "center"
        # flip to match anchor ha settings
        elif label_offset == "right":
            label_ha = "left"
        elif label_offset == "left":
            label_ha = "right"
        self.text = mpl.text.Text(
            self._dist / 2,
            self._deflection,
            label,
            ha=label_ha,
            va="center",
            transform_rotates_text=label_rot,
            rotation=label_flip * 180,
            **label_kw
        )
        # give space for non-centered text
        self.text_shift = np.array([0, 0])
        if not label_rot:
            pad = self.text.get_fontsize() / 2  # in points
        else:
            pad = 0
        if label_ha == "left":
            self.text_shift[0] += pad
        elif label_ha == "right":
            self.text_shift[0] -= pad

    def set_figure(self, figure):

        self.head.set_figure(figure)
        if self.tail:
            self.tail.set_figure(figure)
        self.text.set_figure(figure)
        super().set_figure(figure)

    def set_axes(self, axes):

        self.head.set_axes(axes)
        if self.tail:
            self.tail.set_axes(axes)
        self.text.set_axes(axes)
        super().set_axes(axes)

    def set_transform(self, transform):

        head_transform = (
            affine()
            .scale(self._arrowsize, self._arrowsize)
            .rotate(self._ang)
            .translate(*self._stop)
        )
        self.head.set_transform(head_transform + transform)

        if self.tail:
            tail_transform = (
                affine()
                .scale(self._arrowsize, self._arrowsize)
                .rotate(self._ang + np.pi)
                .translate(*self._start)
            )
            self.tail.set_transform(tail_transform + transform)

        # want to translate text shim in points
        text_shim = affine().translate(*self.text_shift)
        # want main translate in 'data' coords
        text_transform = affine().rotate(self._ang).translate(*self._start)
        self.text.set_transform(text_transform + transform + text_shim)

        # lw_shim = affine().scale(
        line_transform = affine().rotate(self._ang).translate(*self._start)
        super().set_transform(line_transform + transform)

    def draw(self, renderer):

        super().draw(renderer)
        self.head.draw(renderer)
        if self.tail:
            self.tail.draw(renderer)
        self.text.draw(renderer)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import shutil
import importlib.metadata
from pathlib import Path
import nbconvert

# -- Project information (unique to each project) -------------------------------------

project = "leveldiagram"
author = "David Meyer"

release = importlib.metadata.version('leveldiagram')

version = ".".join(release.split('.')[:2])

# HTML icons
img_path = 'img'
html_logo = img_path + '/LevelDiagramIcon.svg'
html_favicon = img_path + '/LevelDiagramIcon.ico'

# -- General configuration (should be identical across all projects) ------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
    "sphinx.ext.mathjax",
]

autosummary_generate = True
autodoc_typehints = 'description'
autoclass_content = 'both'  # options: 'both', 'class', 'init'
numfig = True

# mathjax options for autonumbering equations
mathjax3_config = {'tex': {'tags': 'all', 'useLabelIds': True},}

# Prefix each autosectionlabel with the name of the document it is in and a colon
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The suffix(es) of source filenames.
source_suffix = ['.rst']

# The master toctree document.
master_doc = 'index'

# intersphinx allows us to link directly to other repos sphinxdocs.
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'networkx': ('https://networkx.org/documentation/stable/', None),
}

# Make `some code` equivalent to :code:`some code`
default_role = 'code'

# hide todo notes if on readthedocs and not building the latest
if os.environ.get('READTHEDOCS') and os.environ.get('READTHEDOCS_VERSION') != 'latest':
    todo_include_todos = False
else:
    todo_include_todos = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_title = "leveldiagram"
html_short_title = "leveldiagram"
html_show_copyright = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Customize the html_theme
html_theme_options = {'navigation_depth': 3,
                      'style_nav_header_background':'#607796'}

# -- Options for LaTeX output ---------------------------------------------

latex_engine = 'pdflatex'

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
'papersize': 'a4paper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
'preamble': r'''\usepackage{braket}''',

# Latex figure (float) alignment
'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'leveldiagram.tex', project,
    author,
    'manual')
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = img_path + '/LevelDiagramLogo300.png'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
latex_use_parts = True

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True

def setup(app):

    app.connect('builder-inited', run_nbconvert)


def run_nbconvert(_):
    """
    Runs nbconvert on all example notebooks, for inclusion in docs.
    """

    if os.environ.get('READTHEDOCS'):
        rel_path = '../..'
        imgs_path = os.path.abspath('./img')
    else:
        rel_path = '..'
        imgs_path = os.path.abspath('./source/img')

    examples_path = os.path.join(os.path.abspath(rel_path),'examples')

    out_base_path = os.path.join(os.path.dirname(Path(__file__)),'_examples')
    intro_out_base_path = os.path.join(os.path.dirname(Path(__file__)), '_intro')

    examples = Path(examples_path).glob('*.ipynb')
    intro_nbs = ['Intro']
    img_assets = ['LevelDiagramLogo150.png']

    rst_exporter = nbconvert.exporters.RSTExporter()
    writer = nbconvert.writers.FilesWriter()

    for ex in examples:
        name = ex.stem
        if name in intro_nbs:  
            out_path = intro_out_base_path
        else:
            out_path = os.path.join(out_base_path,name)
        writer.build_directory = out_path
        (body, resources) = rst_exporter.from_filename(ex)
        writer.write(body,resources,name)
        # copy image assets to every directory
        for img in img_assets:
            shutil.copy(os.path.join(imgs_path,img), out_path)



"""
Base Level Diagram class
"""

import matplotlib.pyplot as plt

from typing import Dict, Tuple, Optional, Literal, Any
from networkx import DiGraph
from matplotlib.axes import Axes

from .utils import deep_update, ket_str
from .artists import EnergyLevel, Coupling


class LD:
    """
    Basic Level Diagram drawing class.

    This class is used to draw a level diagram based on a provided Directional Graph.
    The nodes of this graph define the energy levels, the edges define the couplings.

    Note
    ----

    In keeping with the finest matplotlib traditions,
    default options and behavior will produce a *reasonable* output from a graph.
    To get more refined diagrams,
    global options can be set by passing keyword argument
    dictionaries to the constructor.
    Options per level or coupling can be set by setting keyword arguments
    in the dictionaries of the nodes and edges of the graph.

    Examples
    --------
    >>> nodes = (0,1,2)
    >>> edges = ((0,1), (1,2))
    >>> graph = nx.DiGraph()
    >>> graph.add_nodes_from(nodes)
    >>> graph.add_edges_from(edges)
    >>> d = ld.LD(graph)
    >>> d.draw()

    .. image:: img/basic_example.png
        :width: 400
        :alt: Basic 3-level diagram with 2 couplings using all default settings
    """

    _level_defaults = {"width": 1, "color": "k", "text_kw": {"fontsize": "large"}}
    "EnergyLevel default parameters dictionary"

    _coupling_defaults = {"arrowsize": 0.15, "label_kw": {"fontsize": "large"}}
    "Coupling default parameters dictionary"

    _wavy_defaults = {"waveamp": 0.05, "halfperiod": 0.1}
    "Default parameters for a wavy coupling"

    _deflection_defaults = {"deflection": 0.25}
    "Default parameters for a deflection"

    def __init__(
        self,
        graph: DiGraph,
        ax: Optional[Axes] = None,
        default_label: Literal[
            "none", "left_text", "right_text", "top_text", "bottom_text"
        ] = "left_text",
        level_defaults: Optional[Dict[str, Any]] = None,
        coupling_defaults: Optional[Dict[str, Any]] = None,
        wavy_defaults: Optional[Dict[str, Any]] = None,
        deflection_defaults: Optional[Dict[str, Any]] = None,
        use_ld_kw: bool = False,
    ):
        """
        Parameters
        ----------
        graph: networkx.DiGraph
            Graph object that defines the system to diagram

            Beyond the arguments provided to the :class:`Coupling` artist primitive,
            each coupling plotted by :class:`LD` can also take the following parameters
            (which are defined as edge attributes on the graph).

            - **hidden**: bool -
              Tells :class:`LD` to ignore this coupling
            - **start_anchor**: str or 2-element tuple -
              Controls the start anchor point
            - **stop_anchor**: str or 2-element tuple -
              Controls the stop anchor point
            - **detuning**: float -
              How much to detune the coupling from the transition by.
              Defined in x-coordinate units.
            - **wavy**: bool -
              Make coupling arrow a sine wave.
              Uses default options if wavy specific options not provided.
            - **deflect**: bool -
              Make coupling a deflected, circular coupling.
              Uses default options if deflect specific options not provided.

        ax: matplotlib.axes.Axes, optional
            Axes to add the diagram to.
            If None, creates a new figure and axes.
            Default is None.
        default_label: str, optional
            Sets which text label direction to use for default labelling,
            which is the node index inside a key.
            Valid options are `'left_text'`, `'right_text'`,
            `'top_text'`, `'bottom_text'`.
            If 'none', no default labels are not generated.
        level_defaults: dict, optional
            :class:`~.EnergyLevel` default values for whole diagram.
            Provided values override class defaults.
            If None, use class defaults.
        coupling_defaults: dict, optional
            :class:`~.Coupling` default values
            for whole diagram.
            Provided values override class defaults.
            If None, use class defaults.
        wavy_defaults: dict, optional
            Wavy specific :class:`~.Coupling` default values for whole diagram.
            Provided values override class defaults.
            If None, use class defaults.
        deflection_defaults: dict, optional
            Deflection specific :class:`~.Coupling` default values for whole diagram.
            Provided values override class defaults.
            If None, use class defaults.
        """

        if ax is None:
            _, ax = plt.subplots(1)
            ax.set_aspect("equal")
        self.fig = ax.get_figure()
        self.ax = ax

        self.ax.set_axis_off()

        self._graph = graph

        # control parameters
        self.default_label = default_label
        self.use_ld_kw = use_ld_kw

        # save default options for artists
        if level_defaults is None:
            self.level_defaults = self._level_defaults
        else:
            self.level_defaults = deep_update(self._level_defaults, level_defaults)

        if coupling_defaults is None:
            self.coupling_defaults = self._coupling_defaults
        else:
            self.coupling_defaults = deep_update(
                self._coupling_defaults, coupling_defaults
            )

        if wavy_defaults is None:
            self.wavy_defaults = self._wavy_defaults
        else:
            self.wavy_defaults = deep_update(
                self._wavy_defaults, wavy_defaults
            )

        if deflection_defaults is None:
            self.deflection_defaults = self._deflection_defaults
        else:
            self.deflection_defaults = deep_update(
                self._deflection_defaults, deflection_defaults
            )

        # internal storage objects
        self.levels: Dict[int, EnergyLevel] = {}
        """Stores levels to be drawn"""
        self.couplings: Dict[Tuple[int, int], Coupling] = {}
        """Stores couplings to be drawn"""

    def generate_levels(self):
        """
        Creates the EnergyLevel artists from the graph nodes.

        They are saved to the :attr:`levels` dictionary.
        """

        for i, n in enumerate(self._graph.nodes):
            if self.use_ld_kw:
                node = self._graph.nodes[n].get('ld_kw', {}).copy()
            else:
                node = self._graph.nodes[n].copy()
            # if x,y coords not defined, set using node index
            node.setdefault("energy", i)
            node.setdefault("xpos", i)

            if self.default_label != "none":
                node.setdefault(self.default_label, ket_str(n))

            # set default options
            node = deep_update(self.level_defaults, node)
            self.levels[n] = EnergyLevel(**node)

    def generate_couplings(self):
        """
        Creates the Coupling and WavyCoupling artisits from the graph edges.

        They are saved to the :attr:`couplings` dictionary.
        """

        for ed in self._graph.edges:
            if self.use_ld_kw:
                edge = self._graph.edges[ed].get("ld_kw", {}).copy()
            else:
                edge = self._graph.edges[ed].copy()
            # skip if hidden
            if edge.pop("hidden", False):
                continue
            # set default options
            edge = deep_update(self.coupling_defaults, edge)
            # pop off non-arguments
            det = edge.pop("detuning", 0)
            start_anchor = edge.pop("start_anchor", "center")
            stop_anchor = edge.pop("stop_anchor", "center")
            # set where couplings join the levels
            start = self.levels[ed[0]].get_anchor(start_anchor)
            stop = self.levels[ed[1]].get_anchor(stop_anchor)
            # adjust for detuning
            stop[1] -= det
            edge.setdefault("start", start)
            edge.setdefault("stop", stop)
            # auto-cycle colors
            if "color" not in edge:
                edge["color"] = self.ax._get_lines.get_next_color()

            wavy = edge.pop("wavy", False)
            deflect = edge.pop("deflect", False)
            if wavy:
                edge = deep_update(self.wavy_defaults, edge)
            if deflect:
                edge = deep_update(self.deflection_defaults, edge)
                
            self.couplings[ed] = Coupling(**edge)

    def draw(self):
        """
        Add artists to the figure.

        This calls :meth:`matplotlib:matplotlib.axes.Axes.autoscale_view` to ensure
        plot ranges are increased to account for objects.

        It may be necessary to increase plot margins to handle
        labels near edges of the plot.
        """

        self.generate_levels()
        self.generate_couplings()

        for lev in self.levels.values():
            self.ax.add_line(lev)
            for _, text in lev.text_labels.items():
                # registers text labels as artists on the axes
                # ensures text doesn't get clipped by figure edges
                self.ax._add_text(text)

        for coupling in self.couplings.values():
            self.ax.add_line(coupling)

        self.ax.autoscale_view()



"""
Miscellaneous utility functions
"""

import platform
from importlib.metadata import version
from collections.abc import Sequence
import re

from typing import Any

seq_contents = r"[\(|\[\{](.*)[\)|\]\}]"

def ket_str(s: Any) -> str:
    """
    Put a ket around the string in matplotlib.

    If input is a Sequence, but not a str or dict,
    the string representation will only show the container
    contents. For a tuple `(1,2)`, the resulting ket
    will have the form :math:`|1,2\\rangle`.

    Parameters:
        s: Object to be converted to string and placed inside a ket.

    Returns:
        str: A string that will render as :math:`\\left|\\text{s}\\right\\rangle`
    """

    in_s = str(s)

    if not isinstance(s, str) and isinstance(s, Sequence):
        # if sequence, but not string or dict, drop brackets in display
        # use regex to handle more complex things like namedtuples
        in_s = re.search(seq_contents, in_s).group(1)

    out_s = "$\\left|" + in_s + "\\right\\rangle$"

    return out_s


def bra_str(s: Any) -> str:
    """
    Put a bra around the string in matplotlib.

    If input is a Sequence, but not a str or dict,
    the string representation will only show the container
    contents. For a tuple `(1,2)`, the resulting bra
    will have the form :math:`\\langle1,2|`.

    Parameters:
        s: Object to be converted to a string and placed inside a bra.

    Returns:
        str: A string that will render as :math:`\\left\\langle\\text{s}\\right|`
    """

    in_s = str(s)

    if not isinstance(s, str) and isinstance(s, Sequence):
        # if sequence, but not string or dict, drop brackets in display
        # use regex to handle more complex things like namedtuples
        in_s = re.search(seq_contents, in_s).group(1)

    out_s = "$\\left\\langle" + in_s + "\\right|$"

    return out_s


def deep_update(mapping: dict, *updating_mappings: dict) -> dict:
    """
    Helper function to update nested dictionaries.

    Lifted from `pydantic <https://github.com/pydantic/pydantic/blob/main/pydantic/_internal/_utils.py>`_

    Returns:
        dict: Deep-updated copy of `mapping`
    """
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if (
                k in updated_mapping
                and isinstance(updated_mapping[k], dict)
                and isinstance(v, dict)
            ):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping


def about():
    """
    Display version of leveldiagram and critical dependencies.
    """

    header = """
        leveldiagram
    ====================
    """
    print(header)
    print(f'leveldiagram Version: {version("leveldiagram"):s}')
    dep_header = """
        Dependencies
    ====================
    """
    print(dep_header)
    print(f'Python Version:       {platform.python_version():s}')
    print(f'NumPy Version:        {version("numpy"):s}')
    print(f'Matplotlib Version:   {version("matplotlib"):s}')
    print(f'NetworkX Version:     {version("networkx"):s}')
    